#ifndef _FFF_INCLUDE_
#define _FFF_INCLUDE_

/*
    Copyright 2016 Andrew V. Sutherland
    See LICENSE file for license details.
*/

#include "ff.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
    Functional interface to macros in ff.h.
    These inline functions are actually *faster* than macros in many cases (they give the compiler a bit more freedom) and almost never slower.
    This is the recommended interface to the ff library.
*/

static inline void fff_init(long p) { ff_setup_ui(p); }
static inline ff_t fff_char(void) { return _ff_p; }
static inline ff_t fff_zero(void) { return _ff_add_identity; }
static inline ff_t fff_one(void) { return _ff_mult_identity; }
static inline ff_t fff_negone(void) { return _ff_negone; }
static inline ff_t fff_random(void) { ff_t x; ff_random(&x); return x; }

static inline ff_t fff_add(ff_t a, ff_t b) { ff_t c; _ff_add(c,a,b); return c; }
static inline ff_t fff_sub(ff_t a, ff_t b) { ff_t c; _ff_sub(c,a,b); return c; }
static inline ff_t fff_neg(ff_t a) { _ff_neg(a,a); return a; }
static inline ff_t fff_mul(ff_t a, ff_t b) { ff_t c; _ff_mult(c,a,b); return c; }
static inline ff_t fff_sqr(ff_t a) { ff_t c; _ff_square(c,a); return c; }
static inline ff_t fff_dbl(ff_t a) { return fff_add(a,a); }
static inline ff_t fff_haf(ff_t a) { ff_t c; _ff_halve(c,a); return c; }
static inline ff_t fff_dec(ff_t a) { _ff_dec(a); return a; }
static inline ff_t fff_inc(ff_t a) { _ff_inc(a); return a; }
static inline ff_t fff_inv(ff_t a) { ff_t c; _ff_invert(c,a); return c; }
static inline int fff_is_square(ff_t a) { return ff_residue(a); }
static inline int fff_sqrt(ff_t *c,ff_t a) { return ff_sqrt(c,&a); }

static inline ff_t fff_sum_2_products(ff_t a1,ff_t a2, ff_t b2, ff_t b1) // returns a1*b1 + a2*b2
    { ff_t c; _ff_sum_2_mults (c,a1,a2,b2,b1); return c; }
static inline ff_t fff_sum_3_products(ff_t a1,ff_t a2, ff_t a3, ff_t b3, ff_t b2, ff_t b1)
    { ff_t c; _ff_sum_3_mults (c,a1,a2,a3,b3,b2,b1); return c; }
static inline ff_t fff_sum_4_products(ff_t a1,ff_t a2, ff_t a3, ff_t a4, ff_t b4, ff_t b3, ff_t b2, ff_t b1)
    { ff_t c; _ff_sum_4_mults (c,a1,a2,a3,a4,b4,b3,b2,b1); return c; }
static inline ff_t fff_sum_5_products(ff_t a1,ff_t a2, ff_t a3, ff_t a4, ff_t a5, ff_t b5, ff_t b4, ff_t b3, ff_t b2, ff_t b1)
    { ff_t c; _ff_sum_5_mults (c,a1,a2,a3,a4,a5,b5,b4,b3,b2,b1); return c; }
static inline ff_t fff_sum_6_products(ff_t a1,ff_t a2, ff_t a3, ff_t a4, ff_t a5, ff_t a6, ff_t b6, ff_t b5, ff_t b4, ff_t b3, ff_t b2, ff_t b1)
    { ff_t c; _ff_sum_6_mults (c,a1,a2,a3,a4,a5,a6,b6,b5,b4,b3,b2,b1); return c; }

#ifdef __cplusplus
}
#endif

#endif
